"""
Tools for retrieving information about impact-generating nodes in a traversal

collect_stages_by_group(): collect values of stage_property from subfragments of nodes having group_property
"""


from collections import defaultdict, deque


def _is_spec(value):
    if isinstance(value, str):
        if len(value) == 0:
            return False
        return True
    return bool(value)


def _generate_descriptors(_sub_ffs, property):
    for ff in _sub_ffs:
        v = ff.fragment.get(property)
        if _is_spec(v):
            yield v
        else:
            if ff.term.is_process:
                # we reached a terminal node with no matching key
                yield ff.term.term_node.external_ref
            elif ff.term.is_context:
                yield '/'.join(ff.term.term_node.as_list())
            elif ff.term.is_subfrag:
                for v in _generate_descriptors(ff.subfragments, property):
                    yield v
            # nothing to do for cutoffs or foreground fragments


def collect_stages_by_group(_ffs, group_property, stage_property):
    """
    A utility function to report
    :param _ffs:
    :param group_property:
    :param stage_property:
    :return:
    """
    stages = defaultdict(list)
    dq = deque(_ffs)
    while dq:
        ff = dq.popleft()
        g = ff.fragment.get(group_property)
        if _is_spec(g):
            for p in _generate_descriptors([ff], stage_property):
                stages[g].append(p)
        else:
            if ff.term.is_process:
                # we reached a terminal node with no group
                stages[None].append(ff.term.term_node.external_ref)
            elif ff.term.is_context:
                stages[None].append('/'.join(ff.term.term_node.as_list()))
            elif ff.term.is_subfrag:
                dq.extend(ff.subfragments)
            # nothing to do for cutoffs or foreground fragments
    # order-preserving remove duplicates
    return {k: list(dict.fromkeys(v)) for k, v in stages.items()}



from collections import defaultdict
from antelope import ExchangeRef


def frag_flow_lci(fragmentflows, scenario=None):
    """
    Aggregates LCI results recursively into a single set of exchanges.
    The way this works is pretty simple: we go over the traversal. For each node, if the term is a process,
    we grab its LCI, scale each flow by the node weight, and add to the defaultdict
    if the term is a fragment, we call self recursively and add the results to the defaultdict
    if the term is a context, we add it directly to the defaultdict
    if it's a cutoff, we skip
    :param fragmentflows:
    :param scenario:
    :return:
    """
    lci = defaultdict(float)
    parent = None

    for ff in fragmentflows:
        if parent is None:
            # take first one
            parent = ff.fragment
        # _recursive_remote = False
        if ff.term.is_null:
            continue

        node_weight = ff.node_weight
        if node_weight == 0:
            continue

        if ff.term.direction == ff.fragment.direction:
            # if the directions collide (rather than complement), the term is getting run in reverse
            node_weight *= -1

        if ff.term.is_subfrag:
            if len(ff.subfragments) == 0:
                sub_lci = ff.term.term_node.fragment_lci(scenario)  # this does not yet exist
            else:
                sub_lci = frag_flow_lci(ff.subfragments, scenario=ff.subfragment_scenarios)
        else:
            sub_lci = ff.term.unobserved_exchanges()

        for k in sub_lci:
            if k.termination is None:
                '''
                # this is necessary because unobserved_exchanges() returns ExchangeValues for some reason (instead 
                of ExchangeRefs) and they detect and ignore NullContext terminations for some reason, causing 
                null-capped exchanges to show up here as cutoffs.  That may be desirable, actually. maybe we should
                just leave it.  not for now, though.
                '''
                continue
            val = k.value * node_weight
            lci[k.flow, k.direction, k.termination] += val

    output = []
    for k, v in lci.items():
        flow, dirn, term = k
        output.append(ExchangeRef(parent, flow, dirn, value=v, termination=term))

    return sorted(output, key=lambda x: (x.direction == 'Input', x.value), reverse=True)
